/* Copyright (C) 2007 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: ctcss.c,v $
   Revision 1.1  2007/06/30 00:04:25  wb6ymh
   Initial import - source files for thelinkbox.

   
*/ 
#include <inttypes.h>
#include "common.h"
#include "telephony.h"
#include "tone_detect.h"
#include "configvars.h"
#include "ctcss.h"

// Minimum spacing between standard CTCSS tones is 2.3 Hz
#define MIN_CTCSS_SPACING_HZ     2.3

float CTCSS_Freqs[] = {   
   67.0,69.3,71.9,74.4,77.0,79.7,
   82.5,85.4,88.5,91.5,94.8,97.4,
   100.0,103.5,107.2,110.9,114.8,
   118.8,123.0,127.3,131.8,136.5,
   141.3,146.2,151.4,156.7,162.2,
   167.9,173.8,179.9,186.2,192.8,
   203.5,206.5,210.7,218.1,225.7,
   229.1,233.6,241.8,250.3,254.1,0.0
};



static goertzel_descriptor_t LowerDesc;
static goertzel_descriptor_t CenterDesc;
static goertzel_descriptor_t UpperDesc;

static goertzel_state_t LowerState;
static goertzel_state_t CenterState;
static goertzel_state_t UpperState;

static int SamplesProcessed = 0;

// about 150 hz deviation if 5 Khz deviation results in a numeric level 
// of 30000 (91.5 % of maximum)
float DecodeThreshold = 1.3e11;  

int CtcssInit(float Tone)
{
   int i;
   float LowerGuardFreq = 0.0;
   float UpperGuardFreq = 0.0;
   int Ret = 0;

   for(i = 0; CTCSS_Freqs[i] > 0.0; i++) {
      if(Tone == CTCSS_Freqs[i]) {
      // Standard tone
         if(i > 0) {
            LowerGuardFreq = CTCSS_Freqs[i-1];
            UpperGuardFreq = CTCSS_Freqs[i+1];
         }
         break;
      }
   }

   if(LowerGuardFreq == 0.0) {
      LowerGuardFreq = Tone - MIN_CTCSS_SPACING_HZ;
   }

   if(UpperGuardFreq == 0.0) {
      UpperGuardFreq = Tone + MIN_CTCSS_SPACING_HZ;
   }

   make_goertzel_descriptor(&LowerDesc,LowerGuardFreq,CtcssDecoderSamples);
   make_goertzel_descriptor(&CenterDesc,Tone,CtcssDecoderSamples);
   make_goertzel_descriptor(&UpperDesc,UpperGuardFreq,CtcssDecoderSamples);
   goertzel_init(&LowerState,&LowerDesc);
   goertzel_init(&CenterState,&CenterDesc);
   goertzel_init(&UpperState,&UpperDesc);

   return Ret;
}

void UpdateCTCSS(int16 *pPCM,int Samples)
{
   int Samples2Process;
   float LowerGuardVal;
   float CenterVal;
   float UpperGuardVal;
   static float TotalEnergy = 0.0;

   while(Samples > 0) {
      Samples2Process = CtcssDecoderSamples - SamplesProcessed;
      if(Samples2Process > Samples) {
         Samples2Process = Samples;
      }
      goertzel_update(&LowerState,pPCM,Samples2Process);
      goertzel_update(&CenterState,pPCM,Samples2Process);
      goertzel_update(&UpperState,pPCM,Samples2Process);
      // pPCM += Samples2Process;
      SamplesProcessed += Samples2Process;
      Samples -= Samples2Process;

      while(Samples2Process--) {
         TotalEnergy += (*pPCM) * (*pPCM);
         pPCM++;
      }

      if(SamplesProcessed == CtcssDecoderSamples) {
         int bRxCTCSS = g_bRxCTCSS;

         SamplesProcessed = 0;
         LowerGuardVal = goertzel_result(&LowerState);
         CenterVal     = goertzel_result(&CenterState);
         UpperGuardVal = goertzel_result(&UpperState);
         TotalEnergy *= (CtcssDecoderSamples / 2);

         if(CenterVal > DecodeThreshold &&
            CenterVal > LowerGuardVal &&
            CenterVal > UpperGuardVal)
         {  // Our CTCSS tone is present
            g_bRxCTCSS = TRUE;
            if(!bRxCTCSS) {
               printf("%2.2e %2.2e %2.2e %2.2e\n",TotalEnergy,LowerGuardVal,CenterVal,UpperGuardVal);
            }
         }
         else {
         // No tone present
            g_bRxCTCSS = FALSE;
            if(bRxCTCSS) {
               printf("%2.2e %2.2e %2.2e %2.2e",TotalEnergy,LowerGuardVal,CenterVal,UpperGuardVal);
               if(CenterVal < DecodeThreshold) {
                  printf("DecodeThreshold ");
               }

               if(CenterVal < LowerGuardVal) {
                  printf("LowerGuardVal ");
               }

               if(CenterVal < UpperGuardVal) {
                  printf("UpperGuardVal ");
               }
               printf("\n");
            }
         }

         TotalEnergy = 0;
         goertzel_reset(&LowerState);
         goertzel_reset(&CenterState);
         goertzel_reset(&UpperState);
      }
   }
}
