/* Copyright (C) 2007 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: audio.c,v $
   Revision 1.1  2007/06/30 00:04:25  wb6ymh
   Initial import - source files for thelinkbox.

   
*/ 
#include "common.h"

// FreeBSD, Linux, etc..
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <linux/input.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <dirent.h>
#ifdef HAVE_FCNTL_H
   #include <fcntl.h>
#endif

#include "avl.h"
#include "main.h"
#include "conference.h"
#include "audio.h"
#include "ctcss.h"
#include "configvars.h"
#include "devinput.h"

#define OPEN_FLAGS   O_RDONLY | O_NONBLOCK

int InDevFd = -1;
ClientInfo *pInputC = NULL;

static int bCosActive = FALSE;

int DeviceInHandler(ClientInfo *p)
{
   struct input_event Event;
   int Read;
   int bTempCosActiveNotStatic = 2;
   static int bTempCosActive = 2;

   if(p->bReadReady) {
      for( ; ; ) {
         Read = read(p->Socket,&Event,sizeof(Event));
         if(Read <= 0) {
            if(errno != EAGAIN) {
               LOG_ERROR(("%s: read() failed: %s",__FUNCTION__,Err2String(errno)));
            }
            p->bReadReady = FALSE;
            break;
         }
         else if(Read == sizeof(Event)) {
         // Got an event
            LOG_ERROR(("%s: got type: 0x%x, code: 0x%x, value: 0x%x\n",
                       __FUNCTION__,Event.type,Event.code,Event.value));

            if(Event.type == EV_KEY && Event.code == COS_KeyCode) {
            // COS changed state
               if(Event.value) {
                  bTempCosActive = InvertCOS ? FALSE : TRUE;
               }
               else {
                  bTempCosActive = InvertCOS ? TRUE : FALSE;
               }
               bTempCosActiveNotStatic = bTempCosActive;
            }
            else if(Event.type == EV_SYN && bTempCosActive != 2) {
            // Received new COS info
               bCosActive = bTempCosActive;
            }
            else {
               LOG_ERROR(("%s: unhandled\n",__FUNCTION__));
            }

            if(Event.type == EV_SYN && bTempCosActiveNotStatic == 2) {
               LOG_ERROR(("%s: Bug found, bTempCosActiveNotStatic == 2!\n",
                          __FUNCTION__));
            }
         }
         else {
            LOG_ERROR(("%s: read() returned %d bytes, what's up with that?\n",
                       __FUNCTION__));
            p->bReadReady = FALSE;
            break;
         }
      }
   }

   return FALSE;
}

void DeviceInCleanup(ClientInfo *p)
{
   if(p->Socket != -1) {
      close(p->Socket);
   }
}

int InitDeviceInput() 
{
   struct dirent *dirEntry;
   struct stat FileStat;
   char Path[PATH_MAX];
   char DeviceSn[32];
   int Version;
   DIR *dp;
   int fd = -1;
   int i;
   int bDeviceFound = FALSE;
   int Ret = ERR_INPUT_DEV_OPEN; // assume the worse

   if((dp = opendir("/dev/input")) != NULL) {
      while(!bDeviceFound && (dirEntry = readdir(dp)) != NULL) {
         do {
            snprintf(Path,PATH_MAX,"%s/%s","/dev/input",dirEntry->d_name);
            if(stat(Path,&FileStat)) {
            // Couldn't stat the filename, ignore it
               break;
            }

            if(!(FileStat.st_mode & S_IFCHR)) {
            // Not a a character device, ignore it
               break;
            }

            if((fd = open(Path,O_RDONLY)) < 0) {
               LOG_ERROR(("open(%s) failed - %s\n",Path,Err2String(errno)));
               break;
            }

            if(ioctl(fd,EVIOCGVERSION,&Version) < 0) {
               break;
            }

            if((i = ioctl(fd,EVIOCGUNIQ(sizeof(DeviceSn)),DeviceSn)) < 0) {
               break;
            }

            DeviceSn[i] = 0;
            if(strcmp(DeviceSn,AudioDongleSN) == 0) {
               // This is our guy
               bDeviceFound = TRUE;
               LOG_ERROR(("%s: event device for \"%s\" is %s\n",
                          __FUNCTION__,AudioDongleSN,Path));
            }
         } while(FALSE);

         if(fd != -1) {
            close(fd);
            fd = -1;
         }
      }
      closedir(dp);
   }
   else {
      LOG_ERROR(("%s: opendir(/dev/input) failed - %s\n",__FUNCTION__,
                 Err2String(errno)));
   }

   if(bDeviceFound) do {
      pInputC = CreateNewClient();
      if(pInputC == NULL) {
         Ret = ERR_MALLOC;
         break;
      }

      pInputC->Cleanup = DeviceInCleanup;
      pInputC->State = DeviceInHandler;
      if((pInputC->Socket = open(Path,OPEN_FLAGS)) < 0) {
         LOG_ERROR(("%s: open(\"%s\") failed: %s",__FUNCTION__,Path,
                    Err2String(errno)));
         break;
      }

      avl_insert(ClientTree,pInputC);
      pInputC->bInClientTree = TRUE;
      pInputC->bReadWait = TRUE;
      Ret = 0;
   } while(FALSE);

   return Ret;
}

int GetDevInCOS()
{
   return bCosActive;
}
