/* Copyright (C) 2007 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net
   
   $Log$
   
*/ 
#include "common.h"

#include <stdio.h>
#include <string.h>
#include <usb.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "main.h"
#include "usbdev.h"
#include "configvars.h"

#define WRITE_GPIO   1

static usb_dev_handle *udev = NULL;

void UsbKeyRadio(int bKey)
{
   int ret;
   char Buf[2] = {0xff,0xff};

   if(bKey) {
   // Set PTT bit low
      Buf[0] = 0;
   }
   else {
   // Set PTT bit high
      Buf[0] = 0x20;
   }

   ret = usb_control_msg(udev,USB_TYPE_VENDOR,WRITE_GPIO,0,0,Buf,2,1000);
   if(ret != 2) {
      LOG_ERROR(("%s#%d: usb_control_msg returned %d (%s)\n",__FUNCTION__,
                 __LINE__,ret,Err2String(errno)));
   }
}

int FindSoundDev(char *DevAdr)
{
   int i;
   char Path[PATH_MAX];
   char Line[80];
   FILE *fp = NULL;
   char *cp;
   int bDeviceFound = FALSE;

   for(i = 0; i < 32; i++) {
      do {
         sprintf(Path,"/proc/asound/card%d/usbbus",i);
         if((fp = fopen(Path,"r")) == NULL) {
            break;
         }
         if(!fgets(Line,sizeof(Line),fp)) {
            break;
         }
         if((cp = strchr(Line,'\n')) != NULL) {
            *cp = 0;
         }

         if(strcasecmp(Line,DevAdr) == 0) {
            bDeviceFound = TRUE;
         }
      } while(FALSE);

      if(fp != NULL) {
         fclose(fp);
         fp = NULL;
      }

      if(bDeviceFound) {
         break;
      }
   }

   if(i == 32) {
   // return error
      i = -1;
   }

   return i;
}

int UsbInit(void) 
{
   struct usb_bus *bus;
   struct usb_device *dev = NULL;
   int Ret = ERR_USB_DEV_OPEN;   // assume the worse
   int err;
   char DeviceSn[32];
   char DevAdr[32];
   int bDeviceFound = FALSE;

   usb_init();
   usb_find_busses();
   usb_find_devices();

   for(bus = usb_busses; bus && dev == NULL; bus = bus->next) {
      for(dev = bus->devices; dev; dev = dev->next) {
         if(AudioDongleSN != NULL) {
         // Looking for device by SN
            if(dev->descriptor.iSerialNumber) {
               if((udev = usb_open(dev)) == NULL) {
                  LOG_ERROR(("%s#%d: usb_open failed.\n",__FUNCTION__,__LINE__));
               }
               err = usb_get_string_simple(udev,dev->descriptor.iSerialNumber,
                                           DeviceSn,sizeof(DeviceSn));
               usb_close(udev);
               udev = NULL;

               if(err > 0) {
                  if(strcmp(DeviceSn,AudioDongleSN) == 0) {
                     // This is our guy
                     LOG_ERROR(("%s: found USB device \"%s\" @ %s/%s\n",
                                __FUNCTION__,AudioDongleSN,
                                bus->dirname,dev->filename));
                     bDeviceFound = TRUE;
                     break;
                  }
               }
            }
         }
         else if(dev->descriptor.idVendor == 0x074d && 
                 dev->descriptor.idProduct == 0x3556) 
         {
            bDeviceFound = TRUE;
            break;
         }
         else if(dev->descriptor.idVendor == 0x077d && 
                 dev->descriptor.idProduct == 0x07af) 
         {
            bDeviceFound = TRUE;
            break;
         }
      }
      if(bDeviceFound) {
         break;
      }
   }

   if(dev != NULL) {
      if((udev = usb_open(dev)) != NULL) {
         Ret = 0;
      }

   // Find the /dev/dsp entry for this device
      snprintf(DevAdr,sizeof(DevAdr),"%s/%s",bus->dirname,dev->filename);

      if((err = FindSoundDev(DevAdr)) >= 0) {
         char AudioDevicePath[80];
         if(err > 0) {
            snprintf(AudioDevicePath,sizeof(AudioDevicePath),"/dev/dsp%d",err);
         }
         else {
            strcat(AudioDevicePath,"/dev/dsp");
         }
         LOG_ERROR(("%s: USB audio device is %s\n",__FUNCTION__,
                    AudioDevicePath));
         if(AudioDevice != NULL && strcmp(AudioDevice,AudioDevicePath) != 0) {
            LOG_ERROR(("%s: Changing audio device from %s to %s\n",
                       __FUNCTION__,AudioDevice,AudioDevicePath));
         }
         if(AudioDevice != NULL) {
            free(AudioDevice);
         }

         AudioDevice = strdup(AudioDevicePath);
      }
      else {
         LOG_ERROR(("%s: USB audio device file not found\n",__FUNCTION__));
      }
   }
   else {
      LOG_ERROR(("%s: USB device not found\n",__FUNCTION__));
   }

   return Ret;
}

